from django.shortcuts import render
from django.http import HttpResponseRedirect, HttpResponse
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.core.urlresolvers import reverse
from django.core.serializers.python import Serializer
from django.core.serializers.json import DjangoJSONEncoder
from django.views import generic
from django.db.models import Min
import django_filters
import datetime
import json

from birds.models import Animal, Event, Recording
from birds.forms import ClutchForm, BandingForm


class BirdFilter(django_filters.FilterSet):
    class Meta:
        model = Animal
        fields = {
            'sex': ['exact'],
            'band_color': ['exact'],
            'species__code': ['exact'],
        }


class EventFilter(django_filters.FilterSet):
    class Meta:
        model = Event
        fields = {
            'animal__uuid': ['exact','contains'],
            'location__name': ['exact', 'contains'],
            'date': ['exact', 'year', 'range'],
            'entered_by': ['exact'],
        }


class RecordingFilter(django_filters.FilterSet):
    class Meta:
        model = Recording
        fields = {
            'animal__uuid': ['exact', 'contains'],
            'collection__name' : ['exact', 'contains'],
            'datatype__name' : ['exact'],
            'timestamp' : ['exact', 'year', 'range'],
        }


class FlatJsonSerializer(Serializer):
    def get_dump_object(self, obj):
        data = self._current
        if not self.selected_fields or 'id' in self.selected_fields:
            data['id'] = obj.id
        return data

    def end_object(self, obj):
        if not self.first:
            self.stream.write(', ')
        json.dump(self.get_dump_object(obj), self.stream,
                  cls=DjangoJSONEncoder)
        self._current = None

    def start_serialization(self):
        self.stream.write("[")

    def end_serialization(self):
        self.stream.write("]")

    def getvalue(self):
        return super(Serializer, self).getvalue()


def json_response(queryset):
    s = FlatJsonSerializer()
    return HttpResponse(s.serialize(queryset),
                        content_type="application/json")


def bird_list(request, living=None):
    if living:
        qs = Animal.living.annotate(acq_date=Min("event__date")).order_by("acq_date")
    else:
        qs = Animal.objects.all()
    qs = BirdFilter(request.GET, queryset=qs)
    if 'application/json' in request.META.get('HTTP_ACCEPT'):
        return json_response(qs)
    else:
        return render(request, 'birds/birds.html', {'bird_list': qs})


def event_list(request):
    event_list = EventFilter(request.GET, Event.objects.all())
    if 'application/json' in request.META.get('HTTP_ACCEPT'):
        return json_response(event_list)
    else:
        paginator = Paginator(event_list, 25)
        page = request.GET.get('page')
        try:
            qs = paginator.page(page)
        except PageNotAnInteger:
            # If page is not an integer, deliver first page.
            qs = paginator.page(1)
        except EmptyPage:
            # If page is out of range (e.g. 9999), deliver last page of results.
            qs = paginator.page(paginator.num_pages)
        return render(request, 'birds/events.html', {'event_list': qs})


class BirdView(generic.DetailView):
    model = Animal
    template_name = 'birds/bird.html'
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'

    def get_context_data(self, **kwargs):
        context = super(BirdView, self).get_context_data(**kwargs)
        animal = context['animal']
        context['bird_list'] = animal.animal_set.all()
        print(context['bird_list'])
        context['event_list'] = animal.event_set.all()
        return context


class EventListView(generic.ListView):
    template_name = 'birds/events.html'
    context_object_name = 'event_list'
    queryset = Event.objects.order_by('-date')[:100]


class ClutchEntry(generic.FormView):
    template_name = "birds/clutch_entry.html"
    form_class = ClutchForm

    def form_valid(self, form, **kwargs):
        """ For valid entries, render a page with a list of the created events """
        objs = form.create_clutch()
        return render(self.request, 'birds/events.html',
                      { 'event_list': objs['events'],
                        'header_text': 'Hatch events for new clutch'})


class BandingEntry(generic.FormView):
    template_name = "birds/banding_entry.html"
    form_class = BandingForm

    def form_valid(self, form, **kwargs):
        chick = form.create_chick()
        return HttpResponseRedirect(reverse('birds:bird', args=(chick.pk,)))


class IndexView(generic.base.TemplateView):
    template_name = "birds/index.html"

    def get_context_data(self, **kwargs):
        today = datetime.date.today()
        return {"today": today,
                "lastmonth": today.replace(day=1) - datetime.timedelta(days=1)
        }


class EventSummary(generic.base.TemplateView):

    template_name = "birds/summary.html"

    def get_context_data(self, **kwargs):
        from collections import Counter
        tots = Counter()
        year, month = map(int, self.args[:2])
        # aggregation by month does not appear to work properly with postgres
        # backend. Event counts per month will be relatively small, so this
        # shouldn't be too slow
        for event in Event.objects.filter(date__year=year, date__month=month):
            tots[event.status.name] += 1
        return { "year": year,
                 "month": month,
                 "next": datetime.date(year, month, 1) + datetime.timedelta(days=32),
                 "prev": datetime.date(year, month, 1) - datetime.timedelta(days=1),
                 "event_totals": dict(tots) }

# Create your views here.



# -*- coding: utf-8 -*-
# -*- mode: python -*-
from __future__ import unicode_literals

import uuid
from django.core.urlresolvers import reverse
from django.db import models

from django.contrib.auth.models import User
import datetime

import posixpath as pp

class Species(models.Model):
    common_name = models.CharField(max_length=45)
    genus = models.CharField(max_length=45)
    species = models.CharField(max_length=45)
    code = models.CharField(max_length=4)

    def __str__(self):
        return self.common_name

    class Meta:
        ordering = ['common_name']
        verbose_name_plural = 'species'


class Color(models.Model):
    name = models.CharField(max_length=12)
    abbrv = models.CharField('Abbreviation', max_length=3)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']


class Status(models.Model):
    name = models.CharField(max_length=16)
    count = models.SmallIntegerField(default=0, choices=((0, '0'), (-1, '-1'), (1, '+1')),
                                     help_text="1: animal acquired; -1: animal lost; 0: no change")
    category = models.CharField(max_length=2, choices=(('B','B'),('C','C'),('D','D'),('E','E')),
                                blank=True, null=True)
    description = models.TextField()

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']
        verbose_name_plural = 'status codes'


class Location(models.Model):
    name = models.CharField(max_length=45)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']


class Age(models.Model):
    name = models.CharField(max_length=16)
    min_days = models.PositiveIntegerField()
    max_days = models.PositiveIntegerField()
    species = models.ForeignKey('Species')

    def __str__(self):
        return "%s %s (%d-%d days)" % (self.species, self.name, self.min_days, self.max_days)


class LivingAnimalManager(models.Manager):
    def get_queryset(self):
        return super(LivingAnimalManager, self).get_queryset().exclude(event__status__count=-1)


class Animal(models.Model):
    MALE = 'M'
    FEMALE = 'F'
    UNKNOWN_SEX = 'U'
    SEX_CHOICES = (
        (MALE, 'male'),
        (FEMALE, 'female'),
        (UNKNOWN_SEX, 'unknown')
    )

    species = models.ForeignKey('Species')
    sex = models.CharField(max_length=2, choices=SEX_CHOICES)
    band_color = models.ForeignKey('Color', blank=True, null=True)
    band_number = models.IntegerField(blank=True, null=True)
    uuid = models.UUIDField(primary_key=False, default=uuid.uuid4, unique=True)
    parents = models.ManyToManyField('Animal', blank=True)

    reserved_by = models.ForeignKey(User, blank=True, null=True,
                                    help_text="mark a bird as reserved for a specific user")

    created = models.DateTimeField(auto_now_add=True)

    def short_uuid(self):
        return str(self.uuid).split('-')[0]

    def band(self):
        if self.band_number:
            if self.band_color:
                return "%s_%d" % (self.band_color, self.band_number)
            else:
                return "%d" % self.band_number
        else:
            return None

    def name(self):
        return "%s_%s" % (self.species.code, self.band() or self.short_uuid())

    def __str__(self):
        return self.name()

    def sire(self):
        try:
            return self.parents.filter(sex__exact='M')[0]
        except IndexError:
            return None

    def dam(self):
        try:
            return self.parents.filter(sex__exact='F')[0]
        except IndexError:
            return None

    def alive(self):
        """ Returns True if the bird is alive """
        return sum(evt.status.count for evt in self.event_set.all()) > 0

    def nchildren(self):
        """ Returns (living, total) children """
        chicks = self.animal_set
        # probably inefficient
        return (sum(1 for a in chicks.iterator() if a.alive()), chicks.count())

    objects = models.Manager()
    living = LivingAnimalManager()

    def acquisition_event(self):
        """ Returns event when bird was acquired.

        Returns None if no acquisition events
        """
        return self.event_set.filter(status__count=1).order_by('date').first()

    def age_days(self):
        """ Returns days since birthdate or None if unknown"""
        try:
            birthday = self.event_set.filter(status__name="hatched").first().date
            return (datetime.date.today() - birthday).days
        except AttributeError:
            pass

    def get_absolute_url(self):
        return reverse("birds:bird", kwargs={'uuid': self.uuid})

    class Meta:
        ordering = ['band_color', 'band_number']


class Event(models.Model):

    animal = models.ForeignKey('Animal')
    date = models.DateField(default=datetime.date.today)
    status = models.ForeignKey('Status')
    location = models.ForeignKey('Location', blank=True, null=True)
    description = models.TextField(blank=True)

    entered_by = models.ForeignKey(User)
    created = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return "%s: %s on %s" % (self.animal, self.status, self.date)

    def event_date(self):
        """ Description of event and date """
        return "%s on %s" % (self.status, self.date)

    class Meta:
        ordering = ['-date']


class DataCollection(models.Model):
    name = models.CharField(max_length=16, help_text="a short name for the collection")
    uri = models.CharField(max_length=512,
                           help_text="canonical URL for retrieving a recording in this collection")
    def __str__(self):
        return self.name
    class Meta:
        ordering = ['name']


class DataType(models.Model):
    name = models.CharField(max_length=16)
    def __str__(self):
        return self.name
    class Meta:
        ordering = ['name']


class Recording(models.Model):
    animal = models.ForeignKey('Animal')
    collection = models.ForeignKey('DataCollection')
    identifier = models.CharField(max_length=128, help_text="canonical identifier for this recording")

    # optional metadata fields; these will need to be synced with the datafiles
    # somehow, or replaced by external queries
    datatype = models.ForeignKey('DataType', blank=True, null=True)
    timestamp = models.DateTimeField(blank=True, null=True)

    def __str__(self):
        return "%s/%s" % (self.collection.name, self.identifier)

    def uri(self):
        return pp.join(self.collection.uri, self.identifier)

    class Meta:
        ordering = ['animal', 'collection', 'identifier']



from django.contrib.auth.decorators import login_required
from django.conf.urls import url

from birds import views

urlpatterns = [
    url(r'^$', views.IndexView.as_view(), name='index'),
    url(r'^birds/$', views.bird_list, {'living': None}, name='birds'),
    url(r'^birds/living/$', views.bird_list, {'living': True}, name='birds_living'),
    url(r'^birds/(?P<uuid>[a-f0-9\-]{36})/$', views.BirdView.as_view(), name='bird'),
    url(r'^events/$', views.event_list, name='events'),
    url(r'^events/summary/([0-9]{4})/([0-9]{1,2})/$', views.EventSummary.as_view(), name="event_summary"),
    # forms
    url(r'^clutch$', login_required(views.ClutchEntry.as_view()), name='clutch'),
    url(r'^birds/new$', login_required(views.BandingEntry.as_view()), name='new_bird'),
]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('birds', '0003_datacollection_datatype_recording'),
    ]

    operations = [
        migrations.AlterField(
            model_name='datacollection',
            name='name',
            field=models.CharField(help_text='a short name for the collection', max_length=16),
        ),
        migrations.AlterField(
            model_name='datacollection',
            name='uri',
            field=models.CharField(help_text='canonical URL for retrieving a recording in this collection', max_length=512),
        ),
        migrations.AlterField(
            model_name='recording',
            name='identifier',
            field=models.CharField(help_text='canonical identifier for this recording', max_length=128),
        ),
    ]



# -*- coding: utf-8 -*-
# -*- mode: python -*-
from django import forms

from django.contrib.auth.models import User
from birds.models import Animal, Event, Status, Location, Color, Species


class BandingForm(forms.Form):
    acq_status = forms.ModelChoiceField(queryset=Status.objects.filter(count=1))
    acq_date = forms.DateField()
    sire = forms.ModelChoiceField(queryset=Animal.objects.filter(sex__exact='M'),
                                  required=False)
    dam  = forms.ModelChoiceField(queryset=Animal.objects.filter(sex__exact='F'),
                                  required=False)
    species = forms.ModelChoiceField(queryset=Species.objects.all(), required=False)
    banding_date = forms.DateField()
    band_color = forms.ModelChoiceField(queryset=Color.objects.all(), required=False)
    band_number = forms.IntegerField(min_value=1)
    location = forms.ModelChoiceField(queryset=Location.objects.all())
    comments = forms.CharField(widget=forms.Textarea, required=False)
    user = forms.ModelChoiceField(queryset=User.objects.all())

    def clean(self):
        super(BandingForm, self).clean()
        data = self.cleaned_data
        try:
            data['band_status'] = Status.objects.get(name__startswith="band")
        except:
            raise forms.ValidationError("No 'banded' status type - add one in admin")
        if 'acq_status' in data and data['acq_status'].name == "hatched":
            if data['dam'] is None or data['sire'] is None:
                raise forms.ValidationError("Parents required for hatched birds")
            if data['dam'].species != data['sire'].species:
                raise forms.ValidationError("Parents must be the same species")
            data['species'] = data['dam'].species
        else:
            if data['species'] is None:
                raise forms.ValidationError("Species required for non-hatch acquisition")
            data['dam'] = None
            data['sire'] = None
        return data

    def create_chick(self):
        data = self.cleaned_data
        ret = {'chick': None, 'events': []}
        chick = Animal(species=data['species'], sex='U',
                       band_color=data['band_color'], band_number=data['band_number'])
        chick.save()
        if data['sire'] and data['dam']:
            chick.parents.add(data['sire'], data['dam'])
            chick.save()
        evt = Event(animal=chick, date=data['acq_date'],
                    status=data['acq_status'],
                    description=data['comments'],
                    location=data['location'],
                    entered_by=data['user'])
        evt.save()
        evt = Event(animal=chick, date=data['banding_date'],
                    status=data['band_status'],
                    location=data['location'],
                    description=chick.band(),
                    entered_by=data['user'])
        evt.save()
        return chick

class ClutchForm(forms.Form):
    sire = forms.ModelChoiceField(queryset=Animal.objects.filter(sex__exact='M'))
    dam  = forms.ModelChoiceField(queryset=Animal.objects.filter(sex__exact='F'))
    chicks = forms.IntegerField(min_value=1)
    hatch_date = forms.DateField()
    location = forms.ModelChoiceField(queryset=Location.objects.all())
    comments = forms.CharField(widget=forms.Textarea, required=False)
    user = forms.ModelChoiceField(queryset=User.objects.all())

    def clean(self):
        super(ClutchForm, self).clean()
        try:
            self.cleaned_data['status'] = Status.objects.get(name__startswith="hatch")
        except:
            raise forms.ValidationError("No 'hatch' status type - add one in admin")
        if ('dam' in self.cleaned_data and 'sire' in self.cleaned_data and
            self.cleaned_data['dam'].species != self.cleaned_data['sire'].species):
            raise forms.ValidationError("Parents must be the same species")
        return self.cleaned_data


    def create_clutch(self):
        ret = {'chicks': [], 'events': []}
        for i in range(self.cleaned_data['chicks']):
            chick = Animal(species=self.cleaned_data['sire'].species, sex='U')
            chick.save()
            chick.parents.add(self.cleaned_data['sire'],
                              self.cleaned_data['dam'])
            chick.save()
            evt = Event(animal=chick, date=self.cleaned_data['hatch_date'],
                        status=self.cleaned_data['status'],
                        description=self.cleaned_data['comments'],
                        location=self.cleaned_data['location'],
                        entered_by=self.cleaned_data['user'])
            evt.save()
            ret['chicks'].append(chick)
            ret['events'].append(evt)
        return ret



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import uuid
from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('birds', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='animal',
            name='uuid',
            field=models.UUIDField(name='uuid', default=uuid.uuid4),
        ),
    ]



from django.contrib import admin
from birds.models import Species, Color, Location, Animal, Event, Status, Age, DataCollection, DataType, Recording

# Register your models here.

class AnimalAdmin(admin.ModelAdmin):
    fields = ('species', 'sex', 'band_color', 'band_number', 'parents', 'reserved_by')
    list_display = ('name', 'species', 'band', 'uuid', 'sex', 'reserved_by')
    list_filter = ('species', 'sex', 'band_color')
    search_fields = ('band', 'uuid')
    pass


class EventAdmin(admin.ModelAdmin):
    date_hierarchy = 'date'

    fields = ('animal', 'status', 'location', 'description', 'date', 'entered_by')
    list_display = ('animal', 'date', 'status', 'description')
    list_filter = ('animal', 'entered_by', 'status', 'location')
    search_fields = ('animal', 'entered_by', 'description')


class StatusAdmin(admin.ModelAdmin):
    list_display = ('name', 'count')


class RecordingAdmin(admin.ModelAdmin):
    date_hierarchy = 'timestamp'
    list_display = ('animal', 'collection', 'identifier', 'datatype', 'timestamp')
    list_filter = ('animal', 'collection', 'datatype')
    search_fields = ('animal', 'identifier')


admin.site.register(Animal, AnimalAdmin)
admin.site.register(Event, EventAdmin)
admin.site.register(Status, StatusAdmin)
admin.site.register(Recording, RecordingAdmin)

for model in (Species, Color, Location, Age, DataCollection, DataType):
    admin.site.register(model)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('birds', '0002_auto_20150626_1219'),
    ]

    operations = [
        migrations.CreateModel(
            name='DataCollection',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('name', models.CharField(max_length=16)),
                ('uri', models.CharField(max_length=512)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='DataType',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('name', models.CharField(max_length=16)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Recording',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('identifier', models.CharField(max_length=128)),
                ('timestamp', models.DateTimeField(blank=True, null=True)),
                ('animal', models.ForeignKey(to='birds.Animal')),
                ('collection', models.ForeignKey(to='birds.DataCollection')),
                ('datatype', models.ForeignKey(blank=True, to='birds.DataType', null=True)),
            ],
            options={
                'ordering': ['animal', 'collection', 'identifier'],
            },
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import datetime
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Age',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, auto_created=True, verbose_name='ID')),
                ('name', models.CharField(max_length=16)),
                ('min_days', models.PositiveIntegerField()),
                ('max_days', models.PositiveIntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='Animal',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, auto_created=True, verbose_name='ID')),
                ('sex', models.CharField(choices=[('M', 'male'), ('F', 'female'), ('U', 'unknown')], max_length=2)),
                ('band_number', models.IntegerField(blank=True, null=True)),
                ('uuid', models.UUIDField(unique=True, blank=True, editable=False, max_length=32)),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
            options={
                'ordering': ['band_color', 'band_number'],
            },
        ),
        migrations.CreateModel(
            name='Color',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, auto_created=True, verbose_name='ID')),
                ('name', models.CharField(max_length=12)),
                ('abbrv', models.CharField(max_length=3, verbose_name='Abbreviation')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, auto_created=True, verbose_name='ID')),
                ('date', models.DateField(default=datetime.date.today)),
                ('description', models.TextField(blank=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('animal', models.ForeignKey(to='birds.Animal')),
                ('entered_by', models.ForeignKey(to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-date'],
            },
        ),
        migrations.CreateModel(
            name='Location',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, auto_created=True, verbose_name='ID')),
                ('name', models.CharField(max_length=45)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Species',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, auto_created=True, verbose_name='ID')),
                ('common_name', models.CharField(max_length=45)),
                ('genus', models.CharField(max_length=45)),
                ('species', models.CharField(max_length=45)),
                ('code', models.CharField(max_length=4)),
            ],
            options={
                'ordering': ['common_name'],
                'verbose_name_plural': 'species',
            },
        ),
        migrations.CreateModel(
            name='Status',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, auto_created=True, verbose_name='ID')),
                ('name', models.CharField(max_length=16)),
                ('count', models.SmallIntegerField(choices=[(0, '0'), (-1, '-1'), (1, '+1')], help_text='1: animal acquired; -1: animal lost; 0: no change', default=0)),
                ('category', models.CharField(max_length=2, blank=True, choices=[('B', 'B'), ('C', 'C'), ('D', 'D'), ('E', 'E')], null=True)),
                ('description', models.TextField()),
            ],
            options={
                'ordering': ['name'],
                'verbose_name_plural': 'status codes',
            },
        ),
        migrations.AddField(
            model_name='event',
            name='location',
            field=models.ForeignKey(to='birds.Location', null=True, blank=True),
        ),
        migrations.AddField(
            model_name='event',
            name='status',
            field=models.ForeignKey(to='birds.Status'),
        ),
        migrations.AddField(
            model_name='animal',
            name='band_color',
            field=models.ForeignKey(to='birds.Color', null=True, blank=True),
        ),
        migrations.AddField(
            model_name='animal',
            name='parents',
            field=models.ManyToManyField(to='birds.Animal', blank=True),
        ),
        migrations.AddField(
            model_name='animal',
            name='reserved_by',
            field=models.ForeignKey(help_text='mark a bird as reserved for a specific user', null=True, to=settings.AUTH_USER_MODEL, blank=True),
        ),
        migrations.AddField(
            model_name='animal',
            name='species',
            field=models.ForeignKey(to='birds.Species'),
        ),
        migrations.AddField(
            model_name='age',
            name='species',
            field=models.ForeignKey(to='birds.Species'),
        ),
    ]
